#!/usr/bin/env python3
import copy

try:
    import pyrpn
except (ImportError, NameError) as e:
    print("Error importing pyrpn. Try to run make.",
          file=sys.stderr)
    raise e

from PIL import Image
from PIL.PngImagePlugin import PngImageFile, PngInfo
import numpy as np
from tqdm import tqdm

from .ifs1 import IFS1
from .ifs2 import IFS2
from .ifs3 import IFS3
from .ifs4 import IFS4
from .ifs5 import IFS5


#steps = 1024**2
#steps=512**2
steps=1024**2
#steps*=4
#steps*=10
pool_sz = 30
best_sz = 6
n_mutation = 10
outfile='/tmp/rpnifs2_%02d.png'
#world = IFS1.get_world()
#pool = [IFS1(init_sz=3, world=world) for _ in range(pool_sz)]
#world = IFS2.get_world()
#pool = [IFS2(init_sz=3, world=world) for _ in range(pool_sz)]
#world = IFS3.get_world()
#pool = [IFS3(init_sz=3, world=world) for _ in range(pool_sz)]
#world = IFS4.get_world()
#pool = [IFS4(init_sz=3, world=world) for _ in range(pool_sz)]
world = IFS5.get_world()
#pool = [IFS5(init_sz=3, world=world) for _ in range(pool_sz)]
pool = [IFS5(init_sz=6, world=world) for _ in range(pool_sz)]
print('POOL ready')

generation=0
while True:
    generation+=1
    print('='*78)
    print(' '*10 + 'GENERATION #%d' % generation)
    print('='*78)

    scores = []
    i=0
    #for ifs in pool:
    for ifs in tqdm(pool, unit='ifs'):
        i+=1
        ifs.raz_world()
        #for _ in tqdm(range(steps), total=steps, unit_scale=True):
        for _ in range(steps):
            ifs.step()

        score = ifs.score()
        scores.append((ifs, score))
        print('%02d) %5.3f %s' % (i, score, ifs))
        im = Image.fromarray(ifs.get_image())
        im.save(outfile % i)

        im_png = PngImageFile(outfile % i)
        metadata = PngInfo()
        metadata.add_text('Score', '%5.3f' % score)
        metadata.add_text('Generation', '%04d' % generation)
        metadata.add_text('Pool_position', '%02d' % i)
        for key, text in ifs.expr_dict().items():
            metadata.add_text(key, text)
        im_png.save(outfile % i, pnginfo=metadata)

    # Selecting best
    scores = sorted(scores, key=lambda v: 0 if v[1] == np.nan else v[1],
                    reverse=True)
    best = scores[:best_sz]
    for i, (ifs, score) in enumerate(best):
        print('%02d) %5.3f %s' % (i, score, ifs))

    # Mutating best
    pool = [b for b, _ in best]
    for ifs, _ in best:
        for _ in range((pool_sz//best_sz)-1):
            # needs >= IFS5
            while True:
                new = copy.copy(ifs)
                new.mutation(n_mutation)
                for cur in pool:
                    if cur - new == 0:
                        break
                else:
                    pool.append(new)
                    break 
            ###
            #new = copy.copy(ifs)
            #new.mutation(n_mutation)
            #pool.append(new)

